@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2004-2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2004-2006 William Jon McCann <mccann@jhu.edu>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +29,7 @@
 
 #include "gs-manager.h"
 #include "gs-window.h"
+#include "gs-theme-manager.h"
 #include "gs-job.h"
 #include "gs-grab.h"
 #include "gs-fade.h"
@@ -45,6 +46,8 @@ struct GSManagerPrivate
         GSList      *windows;
         GHashTable  *jobs;
 
+        GSThemeManager *theme_manager;
+
         /* Policy */
         glong        lock_timeout;
         glong        cycle_timeout;
@@ -52,29 +55,37 @@ struct GSManagerPrivate
 
         guint        lock_enabled : 1;
         guint        logout_enabled : 1;
+        guint        keyboard_enabled : 1;
         guint        user_switch_enabled : 1;
         guint        throttled : 1;
 
         char        *logout_command;
+        char        *keyboard_command;
 
         /* State */
         guint        active : 1;
         guint        lock_active : 1;
 
+        guint        fading : 1;
         guint        dialog_up : 1;
 
+        time_t       activate_time;
+
         guint        lock_timeout_id;
         guint        cycle_timeout_id;
 
-        time_t       activate_time;
-
         GSList      *themes;
         GSSaverMode  saver_mode;
+        GSGrab      *grab;
+        GSFade      *fade;
+        guint        unfade_idle_id;
 };
 
 enum {
         ACTIVATED,
         DEACTIVATED,
+        AUTH_REQUEST_BEGIN,
+        AUTH_REQUEST_END,
         LAST_SIGNAL
 };
 
@@ -83,17 +94,18 @@ enum {
         PROP_LOCK_ENABLED,
         PROP_LOGOUT_ENABLED,
         PROP_USER_SWITCH_ENABLED,
+        PROP_KEYBOARD_ENABLED,
         PROP_LOCK_TIMEOUT,
         PROP_CYCLE_TIMEOUT,
         PROP_LOGOUT_TIMEOUT,
         PROP_LOGOUT_COMMAND,
+        PROP_KEYBOARD_COMMAND,
         PROP_ACTIVE,
         PROP_THROTTLED,
 };
 
 #define FADE_TIMEOUT 1000
 
-static GObjectClass *parent_class = NULL;
 static guint         signals [LAST_SIGNAL] = { 0, };
 
 G_DEFINE_TYPE (GSManager, gs_manager, G_TYPE_OBJECT)
@@ -135,18 +147,108 @@ select_theme (GSManager *manager)
         return theme;
 }
 
+static GSJob *
+lookup_job_for_window (GSManager *manager,
+                       GSWindow  *window)
+{
+        GSJob *job;
+
+        if (manager->priv->jobs == NULL) {
+                return NULL;
+        }
+
+        job = g_hash_table_lookup (manager->priv->jobs, window);
+
+        return job;
+}
+
 static void
-cycle_job (GSWindow  *window,
-           GSJob     *job,
-           GSManager *manager)
+manager_maybe_stop_job_for_window (GSManager *manager,
+                                   GSWindow  *window)
+{
+        GSJob *job;
+
+        job = lookup_job_for_window (manager, window);
+
+        if (job == NULL) {
+                gs_debug ("Job not found for window");
+                return;
+        }
+
+        gs_job_stop (job);
+}
+
+static void
+manager_maybe_start_job_for_window (GSManager *manager,
+                                    GSWindow  *window)
+{
+        GSJob *job;
+
+        job = lookup_job_for_window (manager, window);
+
+        if (job == NULL) {
+                gs_debug ("Job not found for window");
+                return;
+        }
+
+        if (! manager->priv->dialog_up) {
+                if (! manager->priv->throttled) {
+                        if (! gs_job_is_running (job)) {
+                                if (! gs_window_is_obscured (window)) {
+                                        gs_debug ("Starting job for window");
+                                        gs_job_start (job);
+                                } else {
+                                        gs_debug ("Window is obscured deferring start of job");
+                                }
+                        } else {
+                                gs_debug ("Not starting job because job is running");
+                        }
+                } else {
+                        gs_debug ("Not starting job because throttled");
+                }
+        } else {
+                gs_debug ("Not starting job because dialog is up");
+        }
+}
+
+static void
+manager_select_theme_for_job (GSManager *manager,
+                              GSJob     *job)
 {
         const char *theme;
 
         theme = select_theme (manager);
 
+        if (theme != NULL) {
+                GSThemeInfo    *info;
+                const char     *command;
+
+                command = NULL;
+
+                info = gs_theme_manager_lookup_theme_info (manager->priv->theme_manager, theme);
+                if (info != NULL) {
+                        command = gs_theme_info_get_exec (info);
+                }
+
+                gs_job_set_command (job, command);
+
+
+                if (info != NULL) {
+                        gs_theme_info_unref (info);
+                }
+        } else {
+                gs_job_set_command (job, NULL);
+        }
+}
+
+static void
+cycle_job (GSWindow  *window,
+           GSJob     *job,
+           GSManager *manager)
+{
         gs_job_stop (job);
-        gs_job_set_theme (job, theme, NULL);
-        gs_job_start (job);
+        manager_select_theme_for_job (manager, job);
+        manager_maybe_start_job_for_window (manager, window);
 }
 
 static void
@@ -165,7 +267,7 @@ throttle_job (GSWindow  *window,
         if (manager->priv->throttled) {
                 gs_job_stop (job);
         } else {
-                gs_job_start (job);
+                manager_maybe_start_job_for_window (manager, window);
         }
 }
 
@@ -185,7 +287,7 @@ resume_job (GSWindow  *window,
         if (gs_job_is_running (job)) {
                 gs_job_suspend (job, FALSE);
         } else {
-                gs_job_start (job);
+                manager_maybe_start_job_for_window (manager, window);
         }
 }
 
@@ -232,6 +334,15 @@ gs_manager_set_mode (GSManager  *manager,
         manager->priv->saver_mode = mode;
 }
 
+static void
+free_themes (GSManager *manager)
+{
+        if (manager->priv->themes) {
+                g_slist_foreach (manager->priv->themes, (GFunc)g_free, NULL);
+                g_slist_free (manager->priv->themes);
+        }
+}
+
 void
 gs_manager_set_themes (GSManager *manager,
                        GSList    *themes)
@@ -240,10 +351,7 @@ gs_manager_set_themes (GSManager *manager,
 
         g_return_if_fail (GS_IS_MANAGER (manager));
 
-        if (manager->priv->themes) {
-                g_slist_foreach (manager->priv->themes, (GFunc)g_free, NULL);
-                g_slist_free (manager->priv->themes);
-        }
+        free_themes (manager);
         manager->priv->themes = NULL;
 
         for (l = themes; l; l = l->next) {
@@ -273,12 +381,27 @@ gs_manager_set_throttled (GSManager *manager,
         }
 }
 
+void
+gs_manager_get_lock_active (GSManager *manager,
+                            gboolean  *lock_active)
+{
+        if (lock_active != NULL) {
+                *lock_active = FALSE;
+        }
+
+        g_return_if_fail (GS_IS_MANAGER (manager));
+
+        *lock_active = manager->priv->lock_active;
+}
+
 void
 gs_manager_set_lock_active (GSManager *manager,
                             gboolean   lock_active)
 {
         g_return_if_fail (GS_IS_MANAGER (manager));
 
+        gs_debug ("Setting lock active: %d", lock_active);
+
         if (manager->priv->lock_active != lock_active) {
                 GSList *l;
 
@@ -289,6 +412,19 @@ gs_manager_set_lock_active (GSManager *manager,
         }
 }
 
+void
+gs_manager_get_lock_enabled (GSManager *manager,
+                             gboolean  *lock_enabled)
+{
+        if (lock_enabled != NULL) {
+                *lock_enabled = FALSE;
+        }
+
+        g_return_if_fail (GS_IS_MANAGER (manager));
+
+        *lock_enabled = manager->priv->lock_enabled;
+}
+
 void
 gs_manager_set_lock_enabled (GSManager *manager,
                              gboolean   lock_enabled)
@@ -316,6 +452,22 @@ gs_manager_set_logout_enabled (GSManager *manager,
         }
 }
 
+void
+gs_manager_set_keyboard_enabled (GSManager *manager,
+                                 gboolean   enabled)
+{
+        g_return_if_fail (GS_IS_MANAGER (manager));
+
+        if (manager->priv->keyboard_enabled != enabled) {
+                GSList *l;
+
+                manager->priv->keyboard_enabled = enabled;
+                for (l = manager->priv->windows; l; l = l->next) {
+                        gs_window_set_keyboard_enabled (l->data, enabled);
+                }
+        }
+}
+
 void
 gs_manager_set_user_switch_enabled (GSManager *manager,
                                     gboolean   user_switch_enabled)
@@ -335,8 +487,9 @@ gs_manager_set_user_switch_enabled (GSManager *manager,
 static gboolean
 activate_lock_timeout (GSManager *manager)
 {
-        if (manager->priv->lock_enabled)
+        if (manager->priv->lock_enabled) {
                 gs_manager_set_lock_active (manager, TRUE);
+        }
 
         manager->priv->lock_timeout_id = 0;
 
@@ -408,6 +561,8 @@ void
 gs_manager_set_logout_command (GSManager  *manager,
                                const char *command)
 {
+        GSList *l;
+
         g_return_if_fail (GS_IS_MANAGER (manager));
 
         g_free (manager->priv->logout_command);
@@ -417,6 +572,31 @@ gs_manager_set_logout_command (GSManager  *manager,
         } else {
                 manager->priv->logout_command = NULL;
         }
+
+        for (l = manager->priv->windows; l; l = l->next) {
+                gs_window_set_logout_command (l->data, manager->priv->logout_command);
+        }
+}
+
+void
+gs_manager_set_keyboard_command (GSManager  *manager,
+                                 const char *command)
+{
+        GSList *l;
+
+        g_return_if_fail (GS_IS_MANAGER (manager));
+
+        g_free (manager->priv->keyboard_command);
+
+        if (command) {
+                manager->priv->keyboard_command = g_strdup (command);
+        } else {
+                manager->priv->keyboard_command = NULL;
+        }
+
+        for (l = manager->priv->windows; l; l = l->next) {
+                gs_window_set_logout_command (l->data, manager->priv->keyboard_command);
+        }
 }
 
 gboolean
@@ -425,6 +605,8 @@ gs_manager_cycle (GSManager *manager)
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
+        gs_debug ("cycling jobs");
+
         if (! manager->priv->active) {
                 return FALSE;
         }
@@ -524,6 +706,9 @@ gs_manager_set_property (GObject            *object,
         case PROP_LOGOUT_ENABLED:
                 gs_manager_set_logout_enabled (self, g_value_get_boolean (value));
                 break;
+        case PROP_KEYBOARD_ENABLED:
+                gs_manager_set_keyboard_enabled (self, g_value_get_boolean (value));
+                break;
         case PROP_USER_SWITCH_ENABLED:
                 gs_manager_set_user_switch_enabled (self, g_value_get_boolean (value));
                 break;
@@ -533,6 +718,9 @@ gs_manager_set_property (GObject            *object,
         case PROP_LOGOUT_COMMAND:
                 gs_manager_set_logout_command (self, g_value_get_string (value));
                 break;
+        case PROP_KEYBOARD_COMMAND:
+                gs_manager_set_keyboard_command (self, g_value_get_string (value));
+                break;
         case PROP_CYCLE_TIMEOUT:
                 gs_manager_set_cycle_timeout (self, g_value_get_long (value));
                 break;
@@ -565,6 +753,9 @@ gs_manager_get_property (GObject            *object,
         case PROP_LOGOUT_ENABLED:
                 g_value_set_boolean (value, self->priv->logout_enabled);
                 break;
+        case PROP_KEYBOARD_ENABLED:
+                g_value_set_boolean (value, self->priv->keyboard_enabled);
+                break;
         case PROP_USER_SWITCH_ENABLED:
                 g_value_set_boolean (value, self->priv->user_switch_enabled);
                 break;
@@ -574,6 +765,9 @@ gs_manager_get_property (GObject            *object,
         case PROP_LOGOUT_COMMAND:
                 g_value_set_string (value, self->priv->logout_command);
                 break;
+        case PROP_KEYBOARD_COMMAND:
+                g_value_set_string (value, self->priv->keyboard_command);
+                break;
         case PROP_CYCLE_TIMEOUT:
                 g_value_set_long (value, self->priv->cycle_timeout);
                 break;
@@ -590,8 +784,6 @@ gs_manager_class_init (GSManagerClass *klass)
 {
         GObjectClass   *object_class = G_OBJECT_CLASS (klass);
 
-        parent_class = g_type_class_peek_parent (klass);
-
         object_class->finalize     = gs_manager_finalize;
         object_class->get_property = gs_manager_get_property;
         object_class->set_property = gs_manager_set_property;
@@ -616,6 +808,26 @@ gs_manager_class_init (GSManagerClass *klass)
                               g_cclosure_marshal_VOID__VOID,
                               G_TYPE_NONE,
                               0);
+        signals [AUTH_REQUEST_BEGIN] =
+                g_signal_new ("auth-request-begin",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GSManagerClass, auth_request_begin),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE,
+                              0);
+        signals [AUTH_REQUEST_END] =
+                g_signal_new ("auth-request-end",
+                              G_TYPE_FROM_CLASS (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              G_STRUCT_OFFSET (GSManagerClass, auth_request_end),
+                              NULL,
+                              NULL,
+                              g_cclosure_marshal_VOID__VOID,
+                              G_TYPE_NONE,
+                              0);
 
         g_object_class_install_property (object_class,
                                          PROP_ACTIVE,
@@ -695,6 +907,10 @@ static void
 gs_manager_init (GSManager *manager)
 {
         manager->priv = GS_MANAGER_GET_PRIVATE (manager);
+
+        manager->priv->fade = gs_fade_new ();
+        manager->priv->grab = gs_grab_new ();
+        manager->priv->theme_manager = gs_theme_manager_new ();
 }
 
 static void
@@ -704,6 +920,15 @@ remove_timers (GSManager *manager)
         remove_cycle_timer (manager);
 }
 
+static void
+remove_unfade_idle (GSManager *manager)
+{
+        if (manager->priv->unfade_idle_id > 0) {
+                g_source_remove (manager->priv->unfade_idle_id);
+                manager->priv->unfade_idle_id = 0;
+        }
+}
+
 static void
 gs_manager_finalize (GObject *object)
 {
@@ -717,11 +942,14 @@ gs_manager_finalize (GObject *object)
 
         g_return_if_fail (manager->priv != NULL);
 
+        free_themes (manager);
         g_free (manager->priv->logout_command);
+        g_free (manager->priv->keyboard_command);
 
+        remove_unfade_idle (manager);
         remove_timers (manager);
 
-        gs_grab_release_keyboard_and_mouse ();
+        gs_grab_release (manager->priv->grab);
 
         manager_stop_jobs (manager);
 
@@ -735,7 +963,11 @@ gs_manager_finalize (GObject *object)
         manager->priv->activate_time = 0;
         manager->priv->lock_enabled = FALSE;
 
-        G_OBJECT_CLASS (parent_class)->finalize (object);
+        g_object_unref (manager->priv->fade);
+        g_object_unref (manager->priv->grab);
+        g_object_unref (manager->priv->theme_manager);
+
+        G_OBJECT_CLASS (gs_manager_parent_class)->finalize (object);
 }
 
 static gboolean
@@ -757,7 +989,7 @@ window_deactivated_cb (GSWindow  *window,
 {
         g_return_if_fail (manager != NULL);
         g_return_if_fail (GS_IS_MANAGER (manager));
-        
+
         g_idle_add ((GSourceFunc)window_deactivated_idle, manager);
 }
 
@@ -772,17 +1004,9 @@ window_dialog_up_cb (GSWindow  *window,
 
         gs_debug ("Handling dialog up");
 
+        g_signal_emit (manager, signals [AUTH_REQUEST_BEGIN], 0);
+
         manager->priv->dialog_up = TRUE;
-
-        /* Move keyboard and mouse grabs so dialog can be used */
-        gs_grab_window (gs_window_get_gdk_window (window),
-                        gs_window_get_screen (window),
-                        FALSE);
-
-        /* Release the pointer grab while dialog is up so that
-           the dialog can be used.  We'll regrab it when the dialog goes down. */
-        gs_grab_release_mouse ();
-
         /* Make all other windows insensitive so we don't get events */
         for (l = manager->priv->windows; l; l = l->next) {
                 if (l->data != window) {
@@ -790,6 +1014,16 @@ window_dialog_up_cb (GSWindow  *window,
                 }
         }
 
+        /* Move keyboard and mouse grabs so dialog can be used */
+        gs_grab_move_to_window (manager->priv->grab,
+                                gs_window_get_gdk_window (window),
+                                gs_window_get_screen (window),
+                                FALSE);
+
+        /* Release the pointer grab while dialog is up so that
+           the dialog can be used.  We'll regrab it when the dialog goes down. */
+        gs_grab_release_mouse (manager->priv->grab);
+
         if (! manager->priv->throttled) {
                 gs_debug ("Suspending jobs");
 
@@ -806,43 +1040,63 @@ window_dialog_down_cb (GSWindow  *window,
         g_return_if_fail (manager != NULL);
         g_return_if_fail (GS_IS_MANAGER (manager));
 
+        gs_debug ("Handling dialog down");
+
         /* Regrab the mouse */
-        gs_grab_window (gs_window_get_gdk_window (window),
-                        gs_window_get_screen (window),
-                        FALSE);
+        gs_grab_move_to_window (manager->priv->grab,
+                                gs_window_get_gdk_window (window),
+                                gs_window_get_screen (window),
+                                FALSE);
 
         /* Make all windows sensitive so we get events */
         for (l = manager->priv->windows; l; l = l->next) {
                 gtk_widget_set_sensitive (GTK_WIDGET (l->data), TRUE);
         }
 
+        manager->priv->dialog_up = FALSE;
+
         if (! manager->priv->throttled) {
                 manager_resume_jobs (manager);
         }
 
-        manager->priv->dialog_up = FALSE;
+        g_signal_emit (manager, signals [AUTH_REQUEST_END], 0);
 }
 
-static void
-manager_maybe_start_job_for_window (GSManager *manager,
-                                    GSWindow  *window)
+static GSWindow *
+find_window_at_pointer (GSManager *manager)
 {
-        GSJob *job;
+        GdkDisplay *display;
+        GdkScreen  *screen;
+        int         monitor;
+        int         x, y;
+        GSWindow   *window;
+        int         screen_num;
+        GSList     *l;
 
-        if (manager->priv->jobs == NULL) {
-                return;
+        display = gdk_display_get_default ();
+        gdk_display_get_pointer (display, &screen, &x, &y, NULL);
+        monitor = gdk_screen_get_monitor_at_point (screen, x, y);
+        screen_num = gdk_screen_get_number (screen);
+
+        /* Find the gs-window that is on that screen */
+        window = NULL;
+        for (l = manager->priv->windows; l; l = l->next) {
+                GSWindow *win = GS_WINDOW (l->data);
+                if (gs_window_get_screen (win) == screen
+                    && gs_window_get_monitor (win) == monitor) {
+                        window = win;
+                }
         }
 
-        job = g_hash_table_lookup (manager->priv->jobs, window);
-
-        if (job == NULL) {
-                gs_debug ("Job not found for window");
-                return;
+        if (window == NULL) {
+                gs_debug ("WARNING: Could not find the GSWindow for screen %d", screen_num);
+                /* take the first one */
+                window = manager->priv->windows->data;
+        } else {
+                gs_debug ("Requesting unlock for screen %d", screen_num);
         }
 
-        if (! manager->priv->throttled) {
-                gs_job_start (job);
-        }
+        return window;
 }
 
 static gboolean
@@ -863,9 +1117,11 @@ manager_maybe_grab_window (GSManager *manager,
         grabbed = FALSE;
         if (gs_window_get_screen (window) == screen
             && gs_window_get_monitor (window) == monitor) {
-                gs_grab_window (gs_window_get_gdk_window (window),
-                                gs_window_get_screen (window),
-                                FALSE);
+                gs_debug ("Moving grab to %p", window);
+                gs_grab_move_to_window (manager->priv->grab,
+                                        gs_window_get_gdk_window (window),
+                                        gs_window_get_screen (window),
+                                        FALSE);
                 grabbed = TRUE;
         }
 
@@ -879,12 +1135,29 @@ window_grab_broken_cb (GSWindow           *window,
 {
         gs_debug ("GRAB BROKEN!");
         if (event->keyboard) {
-                gs_grab_keyboard_reset ();
+                gs_grab_keyboard_reset (manager->priv->grab);
         } else {
-                gs_grab_mouse_reset ();
+                gs_grab_mouse_reset (manager->priv->grab);
         }
 }
 
+static gboolean
+unfade_idle (GSManager *manager)
+{
+        gs_debug ("resetting fade");
+        gs_fade_reset (manager->priv->fade);
+        manager->priv->unfade_idle_id = 0;
+        return FALSE;
+}
+
+
+static void
+add_unfade_idle (GSManager *manager)
+{
+        remove_unfade_idle (manager);
+        manager->priv->unfade_idle_id = g_idle_add ((GSourceFunc)unfade_idle, manager);
+}
+
 static gboolean
 window_map_event_cb (GSWindow  *window,
                      GdkEvent  *event,
@@ -917,21 +1190,15 @@ static void
 manager_show_window (GSManager *manager,
                      GSWindow  *window)
 {
-        GSJob      *job;
-        const char *theme;
+        GSJob *job;
 
         job = gs_job_new_for_widget (GTK_WIDGET (window));
 
-        theme = select_theme (manager);
-        gs_job_set_theme (job, theme, NULL);
-
+        manager_select_theme_for_job (manager, job);
         manager_add_job_for_window (manager, window, job);
 
         manager->priv->activate_time = time (NULL);
 
-        /* Set lock state to off initially.  It may be enabled by activate_lock_timout() */
-        gs_manager_set_lock_active (manager, FALSE);
-
         if (manager->priv->lock_timeout >= 0) {
                 remove_lock_timer (manager);
                 add_lock_timer (manager, manager->priv->lock_timeout);
@@ -942,6 +1209,8 @@ manager_show_window (GSManager *manager,
                 add_cycle_timer (manager, manager->priv->cycle_timeout);
         }
 
+        add_unfade_idle (manager);
+
         /* FIXME: only emit signal once */
         g_signal_emit (manager, signals [ACTIVATED], 0);
 }
@@ -960,16 +1229,58 @@ window_show_cb (GSWindow  *window,
         manager_show_window (manager, window);
 }
 
+static void
+maybe_set_window_throttle (GSManager *manager,
+                           GSWindow  *window,
+                           gboolean   throttled)
+{
+        if (throttled) {
+                manager_maybe_stop_job_for_window (manager, window);
+        } else {
+                manager_maybe_start_job_for_window (manager, window);
+        }
+}
+
+static void
+window_obscured_cb (GSWindow   *window,
+                    GParamSpec *pspec,
+                    GSManager  *manager)
+{
+        gboolean obscured;
+
+        obscured = gs_window_is_obscured (window);
+        gs_debug ("Handling window obscured: %s", obscured ? "obscured" : "unobscured");
+
+        maybe_set_window_throttle (manager, window, obscured);
+
+        if (! obscured) {
+                gs_manager_request_unlock (manager);
+        }
+}
+
+static gboolean
+window_activity_cb (GSWindow  *window,
+                    GSManager *manager)
+{
+        gboolean handled;
+
+        handled = gs_manager_request_unlock (manager);
+
+        return handled;
+}
+
 static void
 disconnect_window_signals (GSManager *manager,
                            GSWindow  *window)
 {
         g_signal_handlers_disconnect_by_func (window, window_deactivated_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_activity_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_dialog_up_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_dialog_down_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_show_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_map_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_map_event_cb, manager);
+        g_signal_handlers_disconnect_by_func (window, window_obscured_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_unmap_cb, manager);
         g_signal_handlers_disconnect_by_func (window, window_grab_broken_cb, manager);
 }
@@ -987,6 +1298,8 @@ connect_window_signals (GSManager *manager,
 {
         g_signal_connect_object (window, "destroy",
                                  G_CALLBACK (window_destroyed_cb), manager, 0);
+        g_signal_connect_object (window, "activity",
+                                 G_CALLBACK (window_activity_cb), manager, 0);
         g_signal_connect_object (window, "deactivated",
                                  G_CALLBACK (window_deactivated_cb), manager, 0);
         g_signal_connect_object (window, "dialog-up",
@@ -999,6 +1312,8 @@ connect_window_signals (GSManager *manager,
                                  G_CALLBACK (window_map_cb), manager, G_CONNECT_AFTER);
         g_signal_connect_object (window, "map_event",
                                  G_CALLBACK (window_map_event_cb), manager, G_CONNECT_AFTER);
+        g_signal_connect_object (window, "notify::obscured",
+                                 G_CALLBACK (window_obscured_cb), manager, G_CONNECT_AFTER);
         g_signal_connect_object (window, "unmap",
                                  G_CALLBACK (window_unmap_cb), manager, G_CONNECT_AFTER);
         g_signal_connect_object (window, "grab_broken_event",
@@ -1022,6 +1337,8 @@ gs_manager_create_window (GSManager *manager,
 
         n_monitors = gdk_screen_get_n_monitors (screen);
 
+        gs_debug ("Creating %d windows for screen %d", n_monitors, gdk_screen_get_number (screen));
+
         for (i = 0; i < n_monitors; i++) {
                 window = gs_window_new (screen, i, manager->priv->lock_active);
 
@@ -1029,6 +1346,8 @@ gs_manager_create_window (GSManager *manager,
                 gs_window_set_logout_enabled (window, manager->priv->logout_enabled);
                 gs_window_set_logout_timeout (window, manager->priv->logout_timeout);
                 gs_window_set_logout_command (window, manager->priv->logout_command);
+                gs_window_set_keyboard_enabled (window, manager->priv->keyboard_enabled);
+                gs_window_set_keyboard_command (window, manager->priv->keyboard_command);
 
                 connect_window_signals (manager, window);
 
@@ -1094,31 +1413,35 @@ remove_job (GSJob *job)
         g_object_unref (job);
 }
 
+static void
+fade_done_cb (GSFade    *fade,
+              GSManager *manager)
+{
+        gs_debug ("fade completed, showing windows");
+        show_windows (manager->priv->windows);
+        manager->priv->fading = FALSE;
+}
+
 static gboolean
 gs_manager_activate (GSManager *manager)
 {
-        GdkDisplay *display;
-        GdkScreen  *screen;
-        GdkWindow  *root;
         gboolean    do_fade;
+        gboolean    res;
 
         g_return_val_if_fail (manager != NULL, FALSE);
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
         if (manager->priv->active) {
-                g_warning ("Trying to activate manager when already active");
+                gs_debug ("Trying to activate manager when already active");
                 return FALSE;
         }
 
-        display = gdk_display_get_default ();
-        gdk_display_get_pointer (display, &screen, NULL, NULL, NULL);
-        root = gdk_screen_get_root_window (screen);
-
-        if (! gs_grab_get_keyboard_and_mouse (root, screen)) {
+        res = gs_grab_grab_root (manager->priv->grab, FALSE);
+        if (! res) {
                 return FALSE;
         }
 
-        if (! manager->priv->windows) {
+        if (manager->priv->windows == NULL) {
                 gs_manager_create (GS_MANAGER (manager));
         }
 
@@ -1132,15 +1455,16 @@ gs_manager_activate (GSManager *manager)
         /* fade to black and show windows */
         do_fade = TRUE;
         if (do_fade) {
-                GSFade *fade;
+                manager->priv->fading = TRUE;
+                gs_debug ("fading out");
+                gs_fade_async (manager->priv->fade,
+                               FADE_TIMEOUT,
+                               (GSFadeDoneFunc)fade_done_cb,
+                               manager);
 
-                fade = gs_fade_new ();
-                gs_fade_set_timeout (fade, FADE_TIMEOUT);
-                gs_fade_now (fade);
-                show_windows (manager->priv->windows);
-                g_usleep (500000);
-                gs_fade_reset (fade);
-                g_object_unref (fade);
+                while (manager->priv->fading) {
+                        gtk_main_iteration ();
+                }
         } else {
                 show_windows (manager->priv->windows);
         }
@@ -1157,13 +1481,14 @@ gs_manager_deactivate (GSManager *manager)
         g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
         if (! manager->priv->active) {
-                g_warning ("Trying to deactivate a screensaver that is not active");
+                gs_debug ("Trying to deactivate a screensaver that is not active");
                 return FALSE;
         }
 
+        remove_unfade_idle (manager);
         remove_timers (manager);
 
-        gs_grab_release_keyboard_and_mouse ();
+        gs_grab_release (manager->priv->grab);
 
         manager_stop_jobs (manager);
 
@@ -1173,10 +1498,12 @@ gs_manager_deactivate (GSManager *manager)
         g_slist_free (manager->priv->windows);
         manager->priv->windows = NULL;
 
+        /* reset state */
         manager->priv->active = FALSE;
         manager->priv->activate_time = 0;
         manager->priv->lock_active = FALSE;
-        manager->priv->throttled = FALSE;
+        manager->priv->dialog_up = FALSE;
+        manager->priv->fading = FALSE;
 
         return TRUE;
 }
@@ -1205,32 +1532,46 @@ gs_manager_get_active (GSManager *manager)
         return manager->priv->active;
 }
 
-void
+gboolean
 gs_manager_request_unlock (GSManager *manager)
 {
-        GSList     *l;
-        GdkDisplay *display;
-        GdkScreen  *screen;
-        int         screen_num;
+        GSWindow *window;
 
-        g_return_if_fail (manager != NULL);
-        g_return_if_fail (GS_IS_MANAGER (manager));
+        g_return_val_if_fail (manager != NULL, FALSE);
+        g_return_val_if_fail (GS_IS_MANAGER (manager), FALSE);
 
         if (! manager->priv->active) {
-                return;
+                gs_debug ("Request unlock but manager is not active");
+                return FALSE;
         }
 
-        /* Find the screen that contains the pointer */
-        display = gdk_display_get_default ();
-        gdk_display_get_pointer (display, &screen, NULL, NULL, NULL);
-        screen_num = gdk_screen_get_number (screen);
+        if (manager->priv->dialog_up) {
+                gs_debug ("Request unlock but dialog is already up");
+                return FALSE;
+        }
 
-        /* Find the gs-window that is on that screen */
+        if (manager->priv->fading) {
+                gs_debug ("Request unlock so finishing fade");
+                gs_fade_finish (manager->priv->fade);
+        }
+
+        if (manager->priv->windows == NULL) {
+                gs_debug ("We don't have any windows!");
+                return FALSE;
+        }
+
+        /* Find the GSWindow that contains the pointer */
+        window = find_window_at_pointer (manager);
+        gs_window_request_unlock (window);
+
+        return TRUE;
+}
+
+void
+gs_manager_cancel_unlock_request (GSManager *manager)
+{
+        GSList *l;
         for (l = manager->priv->windows; l; l = l->next) {
-                int num = gdk_screen_get_number (GTK_WINDOW (l->data)->screen);
-                if (num == screen_num) {
-                        gs_window_request_unlock (GS_WINDOW (l->data));
-                        break;
-                }
+                gs_window_cancel_unlock_request (l->data);
         }
 }
